#ifndef __ARCHIVE_PROC_H__
#define __ARCHIVE_PROC_H__

/* 
 * one line to give the Gauche-archive and an idea of what it does.
 * Copyright (C) 2018 aaauuue 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "config.h"

#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* #ifdef HAVE_SYS_TYPES_H */
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* #ifdef HAVE_SYS_STAT_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* #ifdef HAVE_FCNTL_H */

#include <gauche.h>
#include <gauche/extend.h>
#include <gauche/class.h>

#include <archive.h>
#include <archive_entry.h>

#include "archive_class.h"

#define STR(X) #X
#define XSTR(X) STR(X)
#define L __LINE__

#define AR_ERROR(code,str,value) Scm_Values4(SCM_MAKE_INT((code)),SCM_MAKE_STR((str)),SCM_MAKE_STR(__FILE__":"XSTR(L)),(value))
#define COMPARE_SYMBOL(SYM,COMP_SYM_STR) (SCM_EQ((SYM),Scm_MakeSymbol(SCM_STRING(SCM_MAKE_STR((COMP_SYM_STR))),1)))
#define SCM_CSTR(OBJ) (Scm_GetString(SCM_STRING((OBJ))))

int find_slot_index( char *name, ScmObj lst ); /* internal */
ScmObj find_name( struct archive_entry *entry, ScmObj lst );
ScmObj open_read_archive_from_filename( ScmObj obj, ScmObj name );
ScmObj open_read_archive_from_memory( ScmObj obj, ScmObj port );
ScmObj close_read_archive( ScmObj obj );
ScmObj free_read_archive( ScmObj obj );
ScmObj is_read_archive_open( ScmObj obj );
ScmObj get_entry( ScmObj archive, ScmObj entryObj );
ScmObj read_data( ScmObj archive, ScmObj entry );
ScmObj extract_file_to_disk( ScmObj archive, ScmObj entry_list );
ScmObj extract_file_to_memory( ScmObj archive, ScmObj entry_list );
ScmObj extract_all_file( ScmObj archive );
ScmObj create_archive( ScmObj archive_name, ScmObj file_list, ScmObj compress_format, ScmObj filter_format );
ScmObj copy_write_archive( ScmObj tmp_name, ScmObj archive_name, ScmObj remove_file_list, ScmObj append_file_list, ScmObj compress_format, ScmObj filter_format );
ScmObj write_archive_to_disk( ScmObj name, ScmObj file_list, ScmObj compress_format, ScmObj filter_format );
ScmObj write_archive_to_memory( ScmObj size, ScmObj file_list, ScmObj compress_format, ScmObj filter_format );


#endif /* __ARCHIVE_PROC_H__ */
