;;;
;;; Test archive
;;;

(use gauche.test)
(use file.util)
(use rfc.md5)
(use util.digest)
(use gauche.collection)
(use gauche.uvector)

(define *c-src* '("./testdata/org/archive.c" "./testdata/org/archive_class.c"
                  "./testdata/org/archive_proc.c"))
(define *tar1* '("./testdata/org/testarchive01.tar"))
(define *dummy-text* '("./testdata/org/dummy00.txt"))

(define *c-src-md5*
  (map (^x (digest-hexify (md5-digest-string (file->string x))))
        *c-src*))

(define *empty-md5* (list (digest-hexify (md5-digest-string ""))))


(dolist (f (append *c-src* *dummy-text* *tar1* (list "./testdata/org/modify01.tar")))
  (copy-file f (format #f "./testdata/~a" (sys-basename f))))

(test-start "archive")
(use archive)
(use util.archives)
(test-module 'archive)

;; The following is a dummy test code.
;; Replace it for your tests.
(test* "test-archive" "archive is working"
       (test-archive))

(test* "read archive from file" (append *c-src-md5* *empty-md5*)
       (call-with-file-archive
        "./testdata/testarchive01.tar"
        (^a
         (map (^x
               (read-data a x)
               (digest-hexify (md5-digest-string
                               (get-archive-entry-data x)))) a))))

(test* "read archive from memory" (append *c-src-md5* *empty-md5*)
       (let ((oport (open-output-string)))
         (call-with-input-file "./testdata/testarchive01.tar"
           (^(p) (copy-port p oport)))
         (call-with-memory-archive
          oport
          (^a
           (map (^x
                 (read-data a x)
                 (digest-hexify (md5-digest-string
                                 (get-archive-entry-data x)))) a)))))

(test* "create file" '("./testdata/archive.c" "./testdata/archive_class.c" "./testdata/archive_proc.c" "./testdata/dummy00.txt")
       (begin
         (create-archive "./testdata/create00.tar"
                         (list "./testdata/archive.c"
                               "./testdata/archive_class.c"
                               "./testdata/archive_proc.c"
                               "./testdata/dummy00.txt")
                         ARCHIVE-FORMAT-TAR-GNUTAR)
         (call-with-file-archive
          "./testdata/create00.tar"
          (^a
           (map (^x (~ x 'name)) a)))))

(test* "create file with builder" '("./testdata/dummy00.txt" "./testdata/archive_class.c" "./testdata/archive.c" "./testdata/archive_proc.c" "foobardir/")
       (let ((builder (open-build-archive-to-file "./testdata/create01.tar"
                                                  ARCHIVE-FORMAT-TAR-GNUTAR)))
         (dolist (file (list "./testdata/archive.c"
                               "./testdata/archive_class.c"
                               "./testdata/archive_proc.c"
                               "./testdata/dummy00.txt"))
           (let ((e (create-entry-from-file file)))
             (append-entry builder e)
             ;; (describe e)
             ))
         (let ((e (create-entry-from-memory "foobardir/" #f)))
           (set! (~ e 'isdir) #t)
           (append-entry builder e))
         (close-build-archive builder)
         (call-with-file-archive
          "./testdata/create01.tar"
          (^a
           (map (^x (~ x 'name)) a)))))

(test* "create file with builder" '("./testdata/archive_proc.c" "./testdata/archive.c" "./testdata/archive_class.c" "./testdata/dummy00.txt")
       (let ((builder (open-build-archive-to-memory ARCHIVE-FORMAT-TAR-GNUTAR))
             (u8-memory #f))
         (dolist (file (list "./testdata/archive.c"
                               "./testdata/archive_class.c"
                               "./testdata/archive_proc.c"
                               "./testdata/dummy00.txt"))
           (let ((e (create-entry-from-file file)))
             (append-entry builder e)
             ;; (describe e)
             ))
         (set! u8-memory (close-build-archive builder))
         (let ((out (open-output-string)))
           (format out (u8vector->string u8-memory))
           (call-with-memory-archive
            out
            (^a
             (map (^x (~ x 'name)) a))))))

(test* "modify files" #t
       (let ((archive (open-output-archive "./testdata/modify01.tar"
                                           ARCHIVE-FORMAT-TAR-GNUTAR)))
         (append-entry archive (create-entry-from-file "./test.scm"))
         (remove-entry archive (create-entry-from-memory "./testdata/dummy00.txt" (open-output-string)))
         (close-output-archive archive)))

;; If you don't want `gosh' to exit with nonzero status even if
;; the test fails, pass #f to :exit-on-failure.
(test-end :exit-on-failure #t)




