#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

;;; one line to give the Gauche-archive and an idea of what it does.
;;; Copyright (C) 2018 aaauuue 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

(add-load-path ".." :retrive #t)
(define-module util.archives
  (use srfi-35)
  (use util.match)
  (use file.util)
  (use gauche.uvector)
  (use gauche.vport)
  (use gauche.collection)
  (use gauche.mop.instance-pool)

  (use archive)

  (export
   <archive-base>
   <archive-error> 
   <archive-reader-error> <archive-writer-error> <archive-builder-error>
   <archive-entry-error> 
   archive-error?
   archive-error-message archive-error-debug-info
   archive-reader-error?
   archive-reader-error-message archive-reader-error-debug-info
   archive-writer-error?
   archive-writer-error-message archive-writer-error-debug-info
   archive-builder--error?
   archive-builder-error-message archive-builder-error-debug-info
   archive-entry-error?
   archive-entry-error-message archive-entry-error-debug-info
   archive-throw archive-try
   <archive-reader> <archive-writer> <archive-builder> <archive-entry>
   open-input-archive-from-name
   open-input-archive-from-memory
   close-input-archive
   make-archive-entry
   get-archive-entry-data
   flush-archive-entry-data
   create-entry-from-file
   read-file
   create-entry-from-memory
   get-entry
   read-data
   extract-all-files
   extract-files-to-disk
   extract-files-to-memory
   begin-iterator
   end-iterator
   call-with-iterator
   call-with-file-archive
   call-with-memory-archive
   create-archive
   open-output-archive
   close-output-archive
   append-entry
   remove-entry
   modify-entry
   open-build-archive-to-file
   open-build-archive-to-memory
   close-build-archive
   append-entry

   ;;; constant 
   ARCHIVE-FILTER-NONE     
   ARCHIVE-FILTER-GZIP     
   ARCHIVE-FILTER-BZIP2    
   ARCHIVE-FILTER-COMPRESS 
   ARCHIVE-FILTER-PROGRAM  
   ARCHIVE-FILTER-LZMA     
   ARCHIVE-FILTER-XZ       
   ARCHIVE-FILTER-UU       
   ARCHIVE-FILTER-RPM      
   ARCHIVE-FILTER-LZIP     
   ARCHIVE-FILTER-LRZIP   
   ARCHIVE-FILTER-LZOP    
   ARCHIVE-FILTER-GRZIP   
   ARCHIVE-FILTER-LZ4     

   ARCHIVE-FORMAT-BASE-MASK        
   ARCHIVE-FORMAT-CPIO                
   ARCHIVE-FORMAT-CPIO-POSIX          
   ARCHIVE-FORMAT-CPIO-BIN-LE         
   ARCHIVE-FORMAT-CPIO-BIN-BE         
   ARCHIVE-FORMAT-CPIO-SVR4-NOCRC     
   ARCHIVE-FORMAT-CPIO-SVR4-CRC       
   ARCHIVE-FORMAT-CPIO-AFIO-LARGE     
   ARCHIVE-FORMAT-SHAR               
   ARCHIVE-FORMAT-SHAR-BASE          
   ARCHIVE-FORMAT-SHAR-DUMP          
   ARCHIVE-FORMAT-TAR                
   ARCHIVE-FORMAT-TAR-USTAR          
   ARCHIVE-FORMAT-TAR-PAX-INTERCHANGE
   ARCHIVE-FORMAT-TAR-PAX-RESTRICTED 
   ARCHIVE-FORMAT-TAR-GNUTAR         
   ARCHIVE-FORMAT-ISO9660            
   ARCHIVE-FORMAT-ISO9660-ROCKRIDGE  
   ARCHIVE-FORMAT-ZIP                
   ARCHIVE-FORMAT-EMPTY              
   ARCHIVE-FORMAT-AR                 
   ARCHIVE-FORMAT-AR-GNU             
   ARCHIVE-FORMAT-AR-BSD             
   ARCHIVE-FORMAT-MTREE              
   ARCHIVE-FORMAT-RAW                
   ARCHIVE-FORMAT-XAR                
   ARCHIVE-FORMAT-LHA                
   ARCHIVE-FORMAT-CAB                
   ARCHIVE-FORMAT-RAR                
   ARCHIVE-FORMAT-7ZIP               
   ARCHIVE-FORMAT-WARC               


   ))



(select-module util.archives)
(dynamic-load "archive")

(define-condition-type <archive-error> <io-error>
  archive-error?
  (debug-info archive-error-debug-info)
  (message archive-error-message))

(define-condition-type <archive-reader-error> <archive-error> 
  archive-reader-error?
  (debug-info archive-reader-error-debug-info)
  (message archive-reader-error-message))

(define-condition-type <archive-writer-error> <archive-error> 
  archive-writer-error?
  (debug-info archive-writer-error-debug-info)
  (message archive-writer-error-message))

(define-condition-type <archive-builder-error> <archive-error> 
  archive-builder-error?
  (debug-info archive-builder-error-debug-info)
  (message archive-builder-error-message))

(define-condition-type <archive-entry-error> <archive-error> 
  archive-entry-error?
  (debug-info archive-entry-error-debug-info)
  (message archive-entry-error-message))

(define-syntax archive-throw
  (syntax-rules ()
    ((_ (co r d))
     (raise
      (condition
       (co (message r) (debug-info d)))))
    ((_ ((co r d) ...))
     (raise
      (condition
       (co (message r) (debug-info d)) ...)))))

(define-syntax archive-try
  (syntax-rules ()
    ((_ condition-object body ...)
     (receive (code msg dbg data)
         body ...
         (if (not (= code 0))
           (archive-throw (condition-object msg dbg))
           data)))
    ((_ . ?) (syntax-error "malformed syntax "
                           (archive-try condition-object body)))))



(define-class <archive-reader> (<instance-pool-mixin> <collection>)
  ((type :init-form 'file)    ;;; 'file or 'memory
   (name :init-form "")       ;;; archive name
   (data :init-form #f)  
   (archive-core :init-form (make <archive-base>))
   (status :allocation :virtual
           :slot-ref (lambda (o)
                       (receive (code str debug data)
                           (%is-read-archive-open o)
                         (if (= code 0) 'close 'open))))))

(define-class <archive-writer> ()
  ((status :init-form 'close) ;;; 'open or 'close
   (name :init-form "")       ;;; archive name
   (format-type :init-form ARCHIVE-FORMAT-TAR-GNUTAR)
   (filter-type :init-form ARCHIVE-FILTER-NONE)
   (append-list :init-form '())
   (remove-list :init-form '())
   (modify-list :init-form '())))

(define-class <archive-builder> ()
  ((type :init-form 'file) ;; 'file or 'memory
   (name :init-form "")    ;; If type is 'memory always empty.
   (format-type :init-form ARCHIVE-FORMAT-TAR-GNUTAR)
   (filter-type :init-form ARCHIVE-FILTER-NONE)
   (entry-list :init-form '())))

(define-class <archive-entry> (<virtual-output-port>)
  ((isdir :init-form #f)
   (uid :init-form 1000)
   (gid :init-form 1000)
   (mode :init-form 0)
   (atime :init-form 0)
   (ctime :init-form 0)
   (mtime :init-form 0)
   (name :init-form "")
   (size :init-form 0)
   (type :init-form 'file) ;; 'file, 'arcchive or 'memory
   (data :init-form #f)
   (data-opener :init-form (^() #f))
   (data-closer :init-form (^() #f))
   (data-getter :init-form (^() #f))
   (data-flusher :init-form (^(new-out :optional opener closer getter) #f))))



(define (open-input-archive-from-name name)
  (let ((obj (make <archive-reader>)))
    (set! (~ obj 'name) name)
    obj))

(define (open-input-archive-from-memory oport)
  (let ((obj (make <archive-reader>)))
    (set! (~ obj 'type) 'memory)
    (set! (~ obj 'data) oport)
    obj))

(define-method close-input-archive ((cls <archive-reader>))
  (instance-pool-remove! <archive-reader>
                         (^x (eq? x cls)))
  (archive-try <archive-reader-error>
               (%free-read-archive (~ cls 'archive-core))))

(define (make-archive-entry :key (buffer #f) (opener open-output-string) (closer close-output-port) (getter get-output-string) (putb-func write-byte))
  (define (func-setter obj)
    (set! (~ obj 'data-opener) opener)
    (set! (~ obj 'data-closer) closer)
    (set! (~ obj 'data-getter) getter))
  (let* ((out (or buffer (opener)))
         (obj (make <archive-entry>
                :putb (lambda (b) (putb-func b out)))))
    (set! (~ obj 'data) out)
    (func-setter obj)
    (set! (~ obj 'data-flusher)
          (^(new-out :optional (opener open-output-string)
                     (closer close-output-port) (getter get-output-string))
            ((~ obj 'data-closer) out)
            (set! out (or new-out (opener)))
            (set! (~ obj 'data) out)
            (func-setter obj)))
    obj))

(define-method get-archive-entry-data ((cls <archive-entry>))
  (if #0=(~ cls 'data) ((~ cls 'data-getter) #0#) ""))

(define-method flush-archive-entry-data ((cls <archive-entry>) new-out :optional (opener open-output-string) (closer close-output-port) (getter get-output-string))
  (begin0
      (get-archive-entry-data cls)
    ((~ cls 'data-flusher) new-out opener closer getter)))

(define (create-entry-from-file name :optional (opener open-output-string) (closer close-output-port) (getter get-output-string))
  (let ((entry #f)
        (stat #f))
    (if (eq? (file-exists? name) #f)
      (archive-throw
       (<archive-entry-error> #"Can't find ~name" "create-entry-from-file")))
    (set! entry (make-archive-entry :opener opener :closer closer :getter getter))
    (set! stat (sys-stat name))
    (set! (~ entry 'type) 'file)
    (set! (~ entry 'uid) (sys-getuid))
    (set! (~ entry 'gid) (sys-getgid))
    (set! (~ entry 'mode) (~ stat 'mode))
    (set! (~ entry 'atime) (~ stat 'atime))
    (set! (~ entry 'ctime) (~ stat 'ctime))
    (set! (~ entry 'mtime) (~ stat 'mtime))
    (set! (~ entry 'name) name)
    (set! (~ entry 'size) (file-size name))
    entry))

(define (create-entry-from-memory name :optional (buffer #f) (opener open-output-string) (closer close-output-port) (getter get-output-string))
  (let ((entry #f))
    (set! entry (make-archive-entry :buffer buffer :opener opener :closer closer :getter getter))
    (set! (~ entry 'type) 'memory)
    (set! (~ entry 'uid) (sys-getuid))
    (set! (~ entry 'gid) (sys-getgid))
    (set! (~ entry 'mode) 0)
    (set! (~ entry 'atime) (sys-time))
    (set! (~ entry 'ctime) (sys-time))
    (set! (~ entry 'mtime) (sys-time))
    (set! (~ entry 'name) name)
    (set! (~ entry 'size) (if buffer (port-seek buffer 0 SEEK_CUR) 0))
    entry))

(define-method get-entry ((cls <archive-reader>))
  (let ((entry (make-archive-entry)))
    (set! (~ entry 'type) 'archive)
    (archive-try <archive-reader-error>
                 (%get-entry cls entry))
    entry))

(define-method read-file ((cls <archive-entry>))
  (let ((iport #f))
    (if (or (not (eq? (~ cls 'type) 'file))
            (string=? (~ cls 'name) ""))
      (archive-throw
       (<archive-entry-error> "Inappropriate <archive-entry>" "read-file")))
    (begin
      (set! iport (open-input-file (~ cls 'name)))
      (copy-port iport (~ cls 'data))
      (close-port iport))))

(define-method read-data ((cls <archive-reader>) (entry <archive-entry>))
  (if (and (eq? (~ cls 'status) 'close)
           (eq? (~ entry 'type) 'archive))
    (archive-throw
     (<archive-entry-error>
      "This function is available while archive iteration only, thus archive already must be opened."
      "read-data")))
  (archive-try <archive-entry-error>
               (%read-data cls entry)))

(define (re-calc-file-size lst)
  (dolist (entry lst)
    (cond
     ((eq? (~ entry 'type) 'file)
      (set! (~ entry 'size) (file-size (~ entry 'name))))
     ((or (eq? (~ entry 'type) 'memory)
          (eq? (~ entry 'type) 'archive))
      (set! (~ entry 'size) (port-seek (~ entry 'data) 0 SEEK_CUR))))))

(define-method extract-all-files ((cls <archive-reader>))
  (archive-try <archive-reader-error>
               (%extract-all-file cls)))

(define-method extract-files-to-disk ((cls <archive-reader>) entry-list)
  (archive-try <archive-reader-error>
               (%extract-file-to-disk cls)))

(define-method extract-files-to-memory ((cls <archive-reader>) entry-list)
  (archive-try <archive-reader-error>
               (%extract-file-to-memory cls)))

;;; auxiliary method
(define-method begin-iterator ((cls <archive-reader>))
  (cond
   ((eq? (~ cls 'type) 'file)
    (archive-try <archive-reader-error>
                 (%open-read-archive-from-filename cls (~ cls 'name))))
   ((eq? (~ cls 'type) 'memory)
    (archive-try <archive-reader-error>
                 (%open-read-archive-from-memory cls (~ cls 'data))))))

(define-method end-iterator ((cls <archive-reader>))
  (archive-try <archive-reader-error>
               (%close-read-archive cls)))

(define-method call-with-iterator ((cls <archive-reader>) proc . opts)
  (let ((entry #f)
        (ret #t))
    (begin-iterator cls)
    (proc
     (lambda ()
       (set! entry (make-archive-entry))
       (set! (~ entry 'type) 'archive)
       (receive (code msg dbg data)
           (%get-entry cls entry)
         (set! ret (list code msg dbg data)))
       (if (eq? ret #t) #f
           (begin
             (cond
              ((eq? (car ret) #f)
               (end-iterator cls)
               #t)
              ((= (car ret) 0) #f)
              (t
               (match ret
                 ((code msg dbg data)
                  (begin
                    (end-iterator cls)
                    (archive-try <archive-reader-error>
                                 (values code msg dbg data))))))))))
     (lambda () entry))))

(define (call-with-file-archive name proc)
  (let ((archive (open-input-archive-from-name name)))
    (unwind-protect
     (proc archive)
     (close-input-archive archive))))

(define (call-with-memory-archive oport proc)
  (let ((archive (open-input-archive-from-memory oport)))
    (unwind-protect
     (proc archive)
     (close-input-archive archive))))



(define (create-archive archive-name entry-list format-type :optional (filter-type ARCHIVE-FILTER-NONE))
  (archive-try
   <archive-writer-error>
   (%create-archive archive-name entry-list format-type filter-type)))

(define (open-output-archive name format-type :key (filter-type ARCHIVE-FILTER-NONE))
  (let ((writer #f))
    (if (or (eq? (file-exists? name) #f)
            (= (file-size name) 0))
      (archive-throw
       (<archive-writer-error>
        "Archive is not exists or size = 0" "open-output-archive")))
    (set! writer (make <archive-writer>))
    (set! (~ writer 'status) 'open)
    (set! (~ writer 'name) name)
    (set! (~ writer 'format-type) format-type)
    (set! (~ writer 'filter-type) filter-type)
    writer))



(define-method close-output-archive ((cls <archive-writer>))
  (define (if-exists-in-list org dest)
    (not (every (^x (eq? x #f))
                (let ((ret '()))
                  (dolist (v dest (reverse ret))
                    (push! ret
                           (find (^(x)
                                   (and (not (eq? x #f))
                                        (string=? (~ x 'name) (~ v 'name))))
                                 org)))))))
  (define (normalize-list lst)
    (delete-duplicates! lst 
                        (lambda (x y)
                          (string=? (~ x 'name) (~ y 'name))))
    (filter (^x (not (eq? x #f))) lst))
  (define (is-archive-already-allocate? name)
    (find (^x (string=? (~ x 'name) name))
          (instance-pool->list <archive-reader>)))
  (let ((original-list '()) 
        (tmp-name "")
        (ar #f))
    (if (not (eq? (is-archive-already-allocate? (~ cls 'name)) #f))
      (archive-throw
       (<archive-writer-error>
        #"~(~ cls 'name) was already opened. Please close it before writing it."
        "close-output-archive")))
    (set! ar (open-input-archive-from-name (~ cls 'name)))
    (set! original-list (map identity ar))
    (close-input-archive ar)
    (set! #0=(~ cls 'append-list) (normalize-list #0#))
    (set! #1=(~ cls 'remove-list) (normalize-list
                                   (append #1# (~ cls 'modify-list))))
    (set! #2=(~ cls 'modify-list) (normalize-list #2#))
    (if (if-exists-in-list original-list #0#)
      (archive-throw
       ((<archive-entry-error>
         "file for append is already exists in archive."
         "close-output-archive")
        (<archive-writer-error>
         "Can't continue writing archive." "close-output-archive"))))
    (re-calc-file-size (~ cls 'append-list))
    (re-calc-file-size (~ cls 'remove-list))
    (re-calc-file-size (~ cls 'modify-list))

    (set! tmp-name (sys-basename (sys-tmpnam)))
    (%copy-write-archive tmp-name (~ cls 'name)
                         (~ cls 'remove-list)
                         (append (~ cls 'append-list) (~ cls 'modify-list))
                         (~ cls 'format-type) (~ cls 'filter-type))
    (move-file (~ cls 'name)
               (format #f "~a.~a"
                       (~ cls 'name)
                       ($ sys-strftime "%Y%m%d%H%M%S" $ sys-localtime $ sys-time)))
    (move-file tmp-name (~ cls 'name))))


(define-method append-entry ((cls <archive-writer>) entry)
  (set! #0=(~ cls 'append-list) (append (list entry) #0#)))

(define-method remove-entry ((cls <archive-writer>) entry)
  (set! #0=(~ cls 'remove-list) (append (list entry) #0#)))

(define-method modify-entry ((cls <archive-writer>) entry)
  (set! #0=(~ cls 'modify-list) (append (list entry) #0#)))

(define (open-build-archive-to-file name format-type :optional (filter-type ARCHIVE-FILTER-NONE))
  (let ((obj (make <archive-builder>)))
    (set! (~ obj 'name) name)
    (set! (~ obj 'format-type) format-type)
    (set! (~ obj 'filter-type) filter-type)
    obj))

(define (open-build-archive-to-memory format-type :optional (filter-type ARCHIVE-FILTER-NONE))
  (let ((obj (make <archive-builder>)))
    (set! (~ obj 'type) 'memory)
    (set! (~ obj 'format-type) format-type)
    (set! (~ obj 'filter-type) filter-type)
    obj))

(define-method close-build-archive ((cls <archive-builder>))
  (cond
   ((eq? (~ cls 'type) 'file) (close-file-build-archive cls))
   ((eq? (~ cls 'type) 'memory) (close-memory-build-archive cls))
   (t (archive-throw
       (<archive-builder-error> #"Unknown archive type of ~(~ cls 'name)."
                                "close-build-archive")))))

(define (close-file-build-archive archive)
  (re-calc-file-size (~ archive 'entry-list))
  (archive-try <archive-builder-error>
               (%write-archive-to-disk
                (~ archive 'name) (~ archive 'entry-list)
                (~ archive 'format-type) (~ archive 'filter-type))))

(define (close-memory-build-archive archive)
  (let ((size 0))
    (re-calc-file-size (~ archive 'entry-list))
    (set! size (fold (^(x y) (+ (~ x 'size) y))
                     size (~ archive 'entry-list)))
    (archive-try
     <archive-builder-error>
     (%write-archive-to-memory
      size (~ archive 'entry-list)
      (~ archive 'format-type) (~ archive 'filter-type)))))

(define-method append-entry ((cls <archive-builder>) entry)
  (set! #0=(~ cls 'entry-list) (reverse (append (list entry) #0#))))

