
#include "archive_class.h"

SCM_DEFINE_BASE_CLASS( Scm_libarchive_archiveClass, Scm_libarchive_archiveClass,
                       NULL, NULL, NULL, libarchive_archive_allocate, NULL );

ScmObj libarchive_archive_allocate( ScmClass *klass, ScmObj initargs )
{
     Scm_libarchive_archive *p = SCM_NEW( Scm_libarchive_archive );
     SCM_SET_CLASS( p, SCM_CLASS_LIBARCHIVE_ARCHIVE );

     p->obj = NULL;
     p->name = SCM_MAKE_STR( "" );
     p->status = Scm_MakeSymbol( SCM_STRING( SCM_MAKE_STR( "closed" ) ), 1 );
     p->type = GAUCHE_ARCHIVE_NONE;

     return SCM_OBJ( p );
}

void libarchive_archive_finalize( ScmObj obj, void *data )
{
     Scm_libarchive_archive *p = SCM_LIBARCHIVE_ARCHIVE( obj );

     if( p->obj ) {
          switch( p->type ) {
          case GAUCHE_ARCHIVE_NONE:
               archive_read_close( p->obj );
               archive_read_free( p->obj );
          default:
               break;
          }
          p->obj = NULL;
     }
}

ScmObj libarchive_archive_get_name( Scm_libarchive_archive *p )
{
     return p->name;
}

void libarchive_archive_set_name( Scm_libarchive_archive *p, ScmObj value )
{
     if( !SCM_STRINGP( value ) ) {
          Scm_Error( "string required." );
     }

     p->name = value;
}

ScmObj libarchive_archive_get_status( Scm_libarchive_archive *p )
{
     return p->status;
}

void libarchive_archive_set_status( Scm_libarchive_archive *p, ScmObj value )
{
     if( !SCM_SYMBOLP( value ) ) {
          Scm_Error( "symbol required." );
     }

     p->status = value;
}

SCM_DEFINE_BASE_CLASS( Scm_libarchive_archive_entryClass,
                       Scm_libarchive_archive_entryClass, NULL, NULL, NULL,
                       libarchive_archive_entry_allocate, NULL );

ScmObj libarchive_archive_entry_allocate( ScmClass *klass, ScmObj initargs )
{
     Scm_libarchive_archive_entry *p = SCM_NEW( Scm_libarchive_archive_entry );
     SCM_SET_CLASS( p, SCM_CLASS_LIBARCHIVE_ARCHIVE_ENTRY );

     p->archive_name = NULL;
     p->entry_name = NULL;
     p->isdir = SCM_FALSE;
     p->uid = SCM_MAKE_INT( 1000 );
     p->gid = SCM_MAKE_INT( 1000 );
     p->mode = SCM_MAKE_INT( 0 );
     p->atime = SCM_MAKE_INT( 0 );
     p->ctime = SCM_MAKE_INT( 0 );
     p->mtime = SCM_MAKE_INT( 0 );
     p->name = SCM_MAKE_STR( "" );
     p->size = SCM_MAKE_INT( 0 );

     Scm_RegisterFinalizer( SCM_OBJ( p ), libarchive_archive_entry_finalize,
                            NULL );

     return SCM_OBJ( p );
}

void libarchive_archive_entry_finalize( ScmObj obj, void *data )
{
}

ScmObj libarchive_archive_entry_isdir( Scm_libarchive_archive_entry *p )
{
     return p->isdir;
}

ScmObj libarchive_archive_entry_uid( Scm_libarchive_archive_entry *p )
{
     return p->uid;
}

ScmObj libarchive_archive_entry_gid( Scm_libarchive_archive_entry *p )
{
     return p->gid;
}

ScmObj libarchive_archive_entry_mode( Scm_libarchive_archive_entry *p )
{
     return p->mode;
}

ScmObj libarchive_archive_entry_atime( Scm_libarchive_archive_entry *p )
{
     return p->atime;
}

ScmObj libarchive_archive_entry_ctime( Scm_libarchive_archive_entry *p )
{
     return p->ctime;
}

ScmObj libarchive_archive_entry_mtime( Scm_libarchive_archive_entry *p )
{
     return p->mtime;
}

ScmObj libarchive_archive_entry_name( Scm_libarchive_archive_entry *p )
{
     return p->name;
}

ScmObj libarchive_archive_entry_size( Scm_libarchive_archive_entry *p )
{
     return p->size;
}
